// Creating a security descriptor for a new object, a registry key...
#include <windows.h>
#include <stdio.h>
#include <aclapi.h>
 
// Buffer clean up routine
void Cleanup(PSID pEveryoneSID, PSID pAdminSID, PACL pACL, PSECURITY_DESCRIPTOR pSD, HKEY hkSub)
{
	if(pEveryoneSID)
        FreeSid(pEveryoneSID);
    if(pAdminSID)
        FreeSid(pAdminSID);
    if(pACL)
        LocalFree(pACL);
    if(pSD)
        LocalFree(pSD);
    if(hkSub)
        RegCloseKey(hkSub);
}
 
int wmain(int argc, WCHAR **argv)
{
	DWORD dwRes, dwDisposition;
    PSID pEveryoneSID = NULL, pAdminSID = NULL;
    PACL pACL = NULL;
	PSECURITY_DESCRIPTOR pSD = {0};
    // An array of EXPLICIT_ACCESS structure
    EXPLICIT_ACCESS ea[2];
    SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
    SID_IDENTIFIER_AUTHORITY SIDAuthNT = SECURITY_NT_AUTHORITY;
	SECURITY_ATTRIBUTES sa = {0};
    LONG lRes;
    HKEY hkSub = NULL;

    // Create a well-known SID for the Everyone group.
    if(!AllocateAndInitializeSid(&SIDAuthWorld, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0, &pEveryoneSID))
    {
		wprintf(L"AllocateAndInitializeSid() failed, error %u\n", GetLastError());
		Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
        wprintf(L"SID for the Everyone group was allocated and initialized!\n");

    // Initialize an EXPLICIT_ACCESS structure for an ACE.
    // The ACE will allow Everyone read access to the key.
    ZeroMemory(&ea, 2 * sizeof(EXPLICIT_ACCESS));
    ea[0].grfAccessPermissions = KEY_READ;
    ea[0].grfAccessMode = SET_ACCESS;
    ea[0].grfInheritance= NO_INHERITANCE;
    ea[0].Trustee.TrusteeForm = TRUSTEE_IS_SID;
    ea[0].Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
    ea[0].Trustee.ptstrName  = (LPTSTR) pEveryoneSID;

    // Create a SID for the BUILTIN\Administrators group.
    if(!AllocateAndInitializeSid(&SIDAuthNT, 2, SECURITY_BUILTIN_DOMAIN_RID,
		DOMAIN_ALIAS_RID_ADMINS,
		0, 0, 0, 0, 0, 0,
		&pAdminSID))
	{
        wprintf(L"AllocateAndInitializeSid() failed, error %u\n", GetLastError());
        Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
        wprintf(L"SID for the BUILTIN\\Administrators group was allocated and initialized!\n");

    // Initialize an EXPLICIT_ACCESS structure for an ACE.
    // The ACE will allow the Administrators group full access to the key.
    ea[1].grfAccessPermissions = KEY_ALL_ACCESS;
    ea[1].grfAccessMode = SET_ACCESS;
    ea[1].grfInheritance= NO_INHERITANCE;
    ea[1].Trustee.TrusteeForm = TRUSTEE_IS_SID;
    ea[1].Trustee.TrusteeType = TRUSTEE_IS_GROUP;
    ea[1].Trustee.ptstrName  = (LPTSTR) pAdminSID;

    // Create a new ACL that contains the new ACEs.
    dwRes = SetEntriesInAcl(2, ea, NULL, &pACL);
    if(dwRes != ERROR_SUCCESS) 
    {
        wprintf(L"SetEntriesInAcl() failed, error %u\n", GetLastError());
        Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
       wprintf(L"SetEntriesInAcl() for the Administrators group is OK\n");

    // Initialize a security descriptor.
    pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
    if(pSD == NULL) 
    {
        wprintf(L"LocalAlloc() for pSD failed, error %u\n", GetLastError());
        Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
		wprintf(L"LocalAlloc() for pSD is OK!\n");

    if(!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION))
    {
        wprintf(L"InitializeSecurityDescriptor() failed, error %u\n", GetLastError());
        Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
       wprintf(L"InitializeSecurityDescriptor() is pretty fine!\n");

    // Add the ACL to the security descriptor. 
    if(!SetSecurityDescriptorDacl(pSD,
            TRUE,		// bDaclPresent flag 
            pACL, 
            FALSE))		// not a default DACL
    {
        wprintf(L"SetSecurityDescriptorDacl() failed, error %u\n", GetLastError());
        Cleanup(pEveryoneSID, pAdminSID, pACL, pSD, hkSub);
    }
    else
       wprintf(L"SetSecurityDescriptorDacl() is OK!\n");

    // Initialize a security attributes structure.
    sa.nLength = sizeof (SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = pSD;
    sa.bInheritHandle = FALSE;
    // Use the security attributes to set the security descriptor
	// when you create a registry key.
    lRes = RegCreateKeyEx(
		HKEY_CURRENT_USER,	// handle to an open key
		L"MyTestRegKey",	// name of the subkey
		0,					// Reserved
		L"",				// class or object type of this key, may be ignored
		0,					// Options
		KEY_READ | KEY_WRITE, // Access right for the key
		&sa,				// Pointer to security attribute structure,
							// can be inherited or not. NULL is not inherited
		&hkSub,				// variable that receives a handle to the opened or created key
		&dwDisposition);	// variable that receives:
							// REG_CREATED_NEW_KEY - create new key (non-exist)
							// REG_OPENED_EXISTING_KEY - just open the existing key (already exist)
	// If successful
	if(lRes == 0)
	{
		wprintf(L"The value of the \'dwDisposition\': %u\n", dwDisposition);
		wprintf(L"HKEY_CURRENT_USER\\MyTestRegKey successfully created.\n");
	}
	else
		wprintf(L"Creating and opening HKEY_CURRENT_USER\\MyTestRegKey was failed, error %u.\n", GetLastError());

    return 0;
}
